"""
This module contains functionality for accessing the realtime parts of
the PacketLogic system, such as system diagnostics, channel statistics
and dynamic netobject items.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> rt = pl.Realtime()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import verify_ip, negotiate_flicka_with_timeout, deprecatedalias
from thisversion import PL
from thisversion import PLd

import socket
import struct
import select
import time
import sys

from packetlogic2.exceptions import PLDPermissionError, PLDNotAvailable, PLDUnsupportedFeatureError, PLDServerError, PLDUnsupportedInFirmware

class PacketLogicd(object):
	"""
	Class for connecting and communicating with packetlogicd,
	the realtime part of PacketLogic.

	When using the multiversion API you will access this class via the
	L{PLConnection.Realtime() <packetlogic2.PLConnection.Realtime>}
	method.

	All functions that communicate with the PacketLogic Realtime
	system might raise errors common to all or many of the
	functions.Here is a list of such errors, that might not be
	listed on each method:

	@raises ValueError: As the superclass of an internal BadMessageError,
	if a malformed message is recieved on the PacketLogic connection.
	That normally will never happen though.

	@raises Realtime.PLDPermissionError: If the user lacks permission
	        for the requested action.

	@raises IOError: If a generic I/O error occurs while reading or
	        writing on the PacketLogic Realtime connection.

	@group Dynamic NetObject items: dyn_add,dyn_list,dyn_list_full,dyn_list_mask,dyn_list_no,dyn_remove
	@group Firewall Log: fwlog_clear,fwlog_query_offsets,fwlog_query_rule,fwlog_query_time_range
	@group Volume Based Shaping: vbs_list,vbs_query,vbs_reset,vbs_set
	@group Internal: dump_bgptable,dump_drdlstats,inject_packet,reload_drdl
	@group Realtime Data: add_channels_callback,add_netobj_callback,add_service_callback,add_shapingcnt_callback,add_shapingobject_callback,add_sysdiag_callback,add_stats_callback,add_update_done_callback,get_channels_data,get_netobj_data,get_services_data,get_sysdiag_data,get_stats_data,shapingcnt_request_all,stop_updating,update_done_callbacks,update_forever
	
	"""
	
	SUPPORTS_TIMEOUT = True # Actually it is pldb.Ruleset that is checked for this flag, but it seems apropriate to set this too.
	
	def __init__(self, host, user, password, port=42001, timeout=60.0):
		"""
		Initialize and connect to packetlogicd.

		@type host: C{str}
		@param host: A hostname or IP address for the PacketLogic
		             system you want to connect to.

		@type user: C{str}
		@param user: The username you want to authenticate
		             with.  The system share users with the
		             client. You can create the user you want
		             to use with the client.

		@type password: C{str}
		@param password: The password to authenticate with.

		@type port: C{int}
		@param port: The default port is 42001 and normally
		             you don't have to change that.

		@type timeout: C{float} or C{None}
		@param timeout: The connection timeout in seconds. Default None.
			Set to None to use blocking mode during connect, with no specified timeout.
			(A system dependent timeout may or may not still occur after some time.)
			New in version v11.3beta3
		       
		@raises RuntimeError: If authorization fails
		@raises RuntimeError: If server and API version does not match
		@raises IOError: B{"Operation timed out"}: If operation times out without response.
		@raises IOError: If an error occures on the socket when attempting to connect.
		
		"""
		self.host = host
		self.port = port
		self.user = user
		self.timeout = timeout
		#self.timeout = timeout or 60.0 # FIXME Should we set this to the timeout argument too?
		self.password = password
		self.authed = False
		self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.s.setblocking(False)
		self.s.settimeout(timeout)
		try:
			self.s.connect((host, port))
		except socket.error:
			typ, val, tb = sys.exc_info()
			raise IOError, val, tb
		self.io = PL.PLIO(self.s.fileno())
		self.handlers = {}
		self.stats_callbacks = []
		self.chans_callbacks = []
		self.no_callbacks = []
		self.update_done_callbacks = []
		self.service_callbacks = []
		self.shapingobject_callbacks = []
		self.interesting_netobjects = []
		self._update_once_data = None
		self._cache_stats = None
		self._cache_stats_ts = 0
		self._cache_channels = None
		self._cache_channels_ts = 0
		self._NO = RealtimeNetObjectTree()
		self._netobj_ts = 0
		self._features = None
		self._got_statsdata = False

		self._channeldata = {}

		self._hosts = {}

		self._conns = {}

		self._conn_views = []
		self._aggr_views = {}
		self.nextviewid = 1
		self._view_filters = None
		self._view_fields = None

		self._shaping_rule_names = {}

		self._services = {}
		self._service_ts = 0

		self._strtbl = {}

		self._shapingobjects = {}

		negotiate_flicka_with_timeout(self.s, self.io, timeout)

		self.handlers[PLd.PROTOCOL_SERVER_VERSION] = self._handle_vers
		self.handlers[PLd.PROTOCOL_SERVER_AUTHENTICATION_OK] = self._handle_auth
		self.handlers[PLd.PROTOCOL_SERVER_FEATURES] = self._handle_features
		self.handlers[PLd.PROTOCOL_SERVER_EPERM] = self._handle_eperm
		self.handlers[PLd.PROTOCOL_SERVER_NOT_AVAILABLE] = self._handle_eagain
		self.handlers[PLd.PROTOCOL_SERVER_CHANNEL_STATS_INFO] = self._handle_chaninfo
		self.handlers[PLd.PROTOCOL_SERVER_NO_RESET] = self._handle_netobj_reset
		self.handlers[PLd.PROTOCOL_SERVER_UPDATE_DONE] = self._handle_update_done

		self.handlers[PLd.PROTOCOL_SERVER_SHAPING_RULES_INIT] = self._handle_shaping_rules_init

		self._handler_add(PLd.PROTOCOL_SERVER_CONN_UPDATE, self._handle_conn_update)

		self._handler_add(PLd.PROTOCOL_SERVER_HOST_ADD, self._handle_host_add)
		self._handler_add(PLd.PROTOCOL_SERVER_HOST_REMOVE, self._handle_host_remove)
		self._handler_add(PLd.PROTOCOL_SERVER_HOST_UPDATE, self._handle_host_update)
		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_BEGIN, self._handle_strtable_begin)
		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_ENTRIES, self._handle_strtable_entries)
		self._handler_add(PLd.PROTOCOL_SERVER_STRTABLE_END, self._handle_strtable_end)

		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_ADD, self._handle_service_add)
		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_REMOVE, self._handle_service_remove)
		self._handler_add(PLd.PROTOCOL_SERVER_SERVICE_UPDATE, self._handle_service_update)

		self._handler_add(PLd.PROTOCOL_SERVER_SHAPING_OBJECTS_INIT, self._handle_shaping_objects_init)
		self._handler_add(PLd.PROTOCOL_SERVER_SHAPING_OBJECT, self._handle_shaping_object)

		self._handler_add(PLd.PROTOCOL_SERVER_VIEW_ADD_ERROR, self._handle_view_add_error)
		self._handler_add(PLd.PROTOCOL_SERVER_VIEW_UPDATE, self._handle_view_update)
		self._handler_add(PLd.PROTOCOL_SERVER_VIEW_CAPS, self._handle_view_caps)

		PLd.doAuth(self.io, user, password)

		while not self.authed:
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

		# In V11 the server will have NetObjects (realtime stats) visible at start, so send a Hide.
		PLd.netobjectsHide(self.io)

	def _handler_add(self, cmd, handler):
		self.handlers[cmd] = handler

	def _handler_remove(self, cmd):
		if self.handlers.has_key(cmd):
			del self.handlers[cmd]

	def _handle_auth(self, m):
		if PLd.handleAuth(m) == 0:
			raise RuntimeError("Authorization failed")
		self.authed = True

	def _handle_vers(self, m):
		ver, instanceid, systemid, distversion = PLd.handleVersion(m)
		if PLd.PROTOCOL_VERSION == 0:
			import warnings
			warnings.warn("PythonAPI built without protocol version checking. Use at own risk")
		if ver == 0:
			import warnings
			warnings.warn("server built without protocol version checking. Use at own risk")

		if ((PLd.PROTOCOL_VERSION != 0) and
		    (ver != 0) and
		    (ver & ~0x7f) != (PLd.PROTOCOL_VERSION & ~0x7f)):
			raise RuntimeError("Bad server version (got %d expected %d)" % (ver, PLd.PROTOCOL_VERSION))
		self.protocolVersion = ver
		self.instanceid = instanceid
		self.systemid = systemid
		self.distversion = distversion

	def _handle_features(self, m):
		class Features:
			def __repr__(self):
				return "<Features [%s]>" % (", ".join([n for n in dir(self) if not n.startswith('_') and getattr(self, n)]))
			
		bits = PLd.handleFeatures(m)
		features = Features()
		features.bgp = bool(bits & PLd.FEATURES_BGP)
		features.fw = bool(bits & PLd.FEATURES_FW)
		features.shaping = bool(bits & PLd.FEATURES_SHAPING)
		features.statistics = bool(bits & PLd.FEATURES_STATISTICS)
		features.connlog = bool(bits & PLd.FEATURES_CONNLOG)
		features.enterprise = bool(bits & PLd.FEATURES_ENTERPRISE)
		features.monitor = bool(bits & PLd.FEATURES_MONITOR)
		features.divert = bool(bits & PLd.FEATURES_DIVERT)
		features.vbs = bool(bits & PLd.FEATURES_VBS)
		self._features = features

	def _handle_eperm(self, m):
		cmd = PLd.handleEperm(m)
		raise PLDPermissionError("Permission denied executing command (%x)" % cmd)

	def _handle_eagain(self, m):
		raise PLDNotAvailable("Data not available: '%s'" % PLd.handleNotAvail(m))

        def _handle_view_add_error(self, m):
                raise PLDServerError("Error adding view (%d): '%s'" % PLd.handleViewAddError(m))

	def _handle_view_update_add(self, nodes, fields, strings, e):
		id, parent, (level, value), data = e

		if fields[level]['type'] == PLd.PROTOCOL_VIEW_FIELD_TYPE_SERVICE:
			strvalue = self._strtbl.get(value, "<unknown service %d>" % value)
		elif fields[level]['type'] == PLd.PROTOCOL_VIEW_FIELD_TYPE_STRING:
			strvalue = strings.get(value, "<unknown string %d>" % value)
		else:
			strvalue = str(value)

		strvalue = fields[level]['display_prefix'] + strvalue

		nodes[id] = RealtimeAggrNode(strvalue, parent)
		parentnode = nodes[parent]
		parentnode._addchild(nodes[id])
		nodes[id]._set_data(data)

	def _handle_view_update_remove(self, nodes, fields, strings, id):
		node = nodes[id]
		nodes[node.parentid]._removechild(node)
		del nodes[id]

	def _handle_view_update_update(self, nodes, fields, strings, e):
		id, data = e
		nodes[id]._set_data(data)

	def _handle_view_update_string_set(self, nodes, fields, strings, e):
		strid, strdata = e
		strings[strid] = strdata

	def _handle_view_update(self, m):
		viewid = PLd.handleViewUpdateViewID(m)

		nodes, fields, strings, cb = self._aggr_views[viewid]

		fieldtypes = [f['type'] for f in fields]

		entries = PLd.handleViewUpdate(m, fieldtypes)

		handlers = {PLd.PROTOCOL_SERVER_VIEW_UPDATE_OP_ADD: self._handle_view_update_add,
			    PLd.PROTOCOL_SERVER_VIEW_UPDATE_OP_REMOVE: self._handle_view_update_remove,
			    PLd.PROTOCOL_SERVER_VIEW_UPDATE_OP_UPDATE: self._handle_view_update_update,
			    PLd.PROTOCOL_SERVER_VIEW_UPDATE_OP_STRING_SET: self._handle_view_update_string_set}

		for op, data in entries:
			handlers[op](nodes, fields, strings, data)

	def _handle_view_caps(self, m):
		self._view_filters, self._view_fields = PLd.handleViewCaps(m)

	def _handle_chaninfo(self, m):
		self._channeldata = {}
		chans = PLd.handleChannelsInfo(m)
		def datahandler(m):
			r = PLd.handleChannelsData(m, chans)
			self._channeldata.update(r)

		self._handler_add(PLd.PROTOCOL_SERVER_CHANNEL_STATS_DATA, datahandler)

	def _has_feature(self, feature):
		""" Return True or False depending on if the feature is enabled."""
		if self._features is None:
			raise RuntimeError, "Never recieved features message at connect."

		return getattr(self._features, feature)
		
	def _require_features(self, *features):
		""" If any of the listed features aren't enabled, a PLDUnsupportedFeatureError is raised."""
		for feature in features:
			if not self._has_feature(feature):
				raise PLDUnsupportedFeatureError
				

	def _fullflush(self):
		"""Make sure any outstanding messages are sent and
		   handled by the server by sending a ping request and
		   waiting for the response, which makes sure that all
		   messages sent before this are handled by server"""
		PLd.ping(self.io)
		self._do(None, PLd.PROTOCOL_SERVER_PONG)

	def _flush(self):
		while self.io.wantWrite():
			outp = [self.s]

			(i, o, e) = select.select([], outp, outp, self.timeout)

			if len(i) + len(o) + len(e) == 0:
				raise IOError("Operation timed out")

			if len(e) > 0:
				raise IOError("Select returned error")
			
			if len(o) > 0:
				if not self.io.doWrite():
					raise IOError("Error writing to socket")

	def _handle_update_done(self, m):
		for f,under,include_hosts in self.no_callbacks:
			if not self.handlers.has_key(PLd.PROTOCOL_SERVER_NO_UPDATE):
				self._NO._link_parents()
				self._handler_add(PLd.PROTOCOL_SERVER_NO_UPDATE, self._handle_netobj_updateinfo)

			if under is not None:
				underno = self._NO.find(under[:-1])
				if underno is None:
					f(_RealtimeNetObjectList())
				else:
					l = _RealtimeNetObjectList(self._NO.list(under[:-1], recursive=False))
					if include_hosts:
						noid = underno.id
						l._hosts_func = lambda :[h for h in self._hosts.itervalues() if noid in h.netobjects]
					f(l)
			else:
				f(self._NO)
		if self._got_statsdata:
			self._got_statsdata = False
			czones = dict(map(lambda q: (q[0], dict(map(lambda a: (a['name'], a), q[1]))), self.sysdiag_zones.values()))
			for f in self.stats_callbacks:
				f(czones)

		for f in self.service_callbacks:
			f(self._services)

		for f in self.shapingobject_callbacks:
			f(self._shapingobjects)

		for f in self.chans_callbacks:
			f(self._channeldata)

		for viewid, f in self._conn_views:
			conns = [c for c in self._conns.itervalues() if c._in_view(viewid)]
			for c in conns:
				c._service = self._getstrtblentry(c._service_id)
				c._base_service = self._getstrtblentry(c._base_service_id)
				c._shaping_rules = map(self._resolve_shaping_rule_name, c._shaping_rule_ids)
			f(conns)

		for nodes, _, _, cb in self._aggr_views.itervalues():
			cb(nodes[0])

		for c in self.update_done_callbacks:
			c()

	def _resolve_shaping_rule_name(self, idx):
		return self._shaping_rule_names.get(idx, "Rule not in table? (%d)" % idx)

	def _handle_strtable_begin(self, m):
		s, l = PLd.handleStrtblBegin(m)
		self._loading_strtbl_siz = s
		self._loading_strtbl = l
	def _handle_strtable_entries(self, m):
		self._loading_strtbl += PLd.handleStrtblEntries(m)
	def _handle_strtable_end(self, m):
		assert len(self._loading_strtbl) == self._loading_strtbl_siz
		self._strtbl = dict(self._loading_strtbl)
		self._loading_strtbl_siz = None
		self._loading_strtbl = None

	def _handle_shaping_rules_init(self, m):
		self._shaping_rule_names = PLd.handleShapingRulesInit(m)

	def _handle_netobj_reset(self, m):
		"""
		This is an internal function which converts NetObject Reset messages
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			r = PLd.handleNOReset(m)
			self._NO._clear()
			self._hosts.clear()

		self._handler_add(PLd.PROTOCOL_SERVER_NO_RESET, datahandler)
		self._handler_remove(PLd.PROTOCOL_SERVER_NO_UPDATE)
		self.handlers[PLd.PROTOCOL_SERVER_NO_RESET](m)

	def _handle_netobj_add(self, m):
		"""
		This is an internal function which converts NetObject Add messages
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			nolist = PLd.handleNOAdd(m, RealtimeNetObject)
			self._NO._extend(nolist)
			for no in nolist:
				self._NO._link_parent_one(no)
				for noname in self.interesting_netobjects:
					fp = no.fullpath + "/"
					if noname.startswith(fp):
						PLd.netobjectShow(self.io, no.id)
						break

		self._handler_add(PLd.PROTOCOL_SERVER_NO_ADD, datahandler)
		self.handlers[PLd.PROTOCOL_SERVER_NO_ADD](m)

	def _handle_netobj_remove(self, m):
		self._NO._remove(PLd.handleNORemove(m))

	def _handle_netobj_updateinfo(self, m):
		"""
		This is an internal function which converts NetObject Update messges
		into Python data to be used by the NetObject callback function
		"""
		def datahandler(m):
			PLd.handleNOUpdate(m, self._NO._by_id)
					
		self._handler_add(PLd.PROTOCOL_SERVER_NO_UPDATE, datahandler)
		self.handlers[PLd.PROTOCOL_SERVER_NO_UPDATE](m)

	def _getstrtblentry(self, strid):
		return self._strtbl.get(strid, "Service not in stringtable (%d)" % strid)

	def _handle_service_dict(self, r):
		for k,v in r.iteritems():
			sn = self._getstrtblentry(k)
			if sn not in self._services:
				self._services[sn] = {}
			d = self._services[sn]
			d.update(v)
	def _handle_service_add(self, m):
		r = PLd.handleServiceAdd(m)
		self._handle_service_dict(r)
	def _handle_service_remove(self, m):
		r = PLd.handleServiceRemove(m)
	def _handle_service_update(self, m):
		r = PLd.handleServiceUpdate(m)
		self._handle_service_dict(r)

	def _handle_host_add(self, m):
		PLd.handleHostAdd(m, RealtimeHost, self._hosts)

	def _handle_host_remove(self, m):
		hid = PLd.handleHostRemove(m)
		if hid not in self._hosts:
			print "Strange, remove of nonexistant host"
		del self._hosts[hid]

	def _handle_host_update(self, m):
		PLd.handleHostUpdate(m, self._hosts)

	def _handle_conn_update(self, m):
		PLd.handleConnUpdate(m, self._conns, RealtimeConn)

	def _handle_shaping_objects_init(self, m):
		self._shapingobjects = {}
	def _handle_shaping_object(self, m):
		PLd.shapingobjectUpdateHandle(m, self._shapingobjects)

	def _fill_socketlist(self, inl, outl):
		inl.append(self.s)
		if self.io.wantWrite():
			outl.append(self.s)

	def _handle_socket_events(self, i, o, e):
		if self.s in e:
			""" Huston! We have a problem! """
			raise IOError("Select returned error")

		if self.s in o:
			if not self.io.doWrite():
				raise IOError("Error writing to socket")

		if self.s in i:
			if not self.io.doRead():
				raise IOError("Error reading from socket")

			m = self.io.readMsg()
			while m:
				cmd = m.cmd()
				if self.handlers.has_key(cmd):
					if self.handlers[cmd](m):
						break

				m = self.io.readMsg()


	def _process(self, timeout):
		"""
		This is an internal function that will process events as they
		occur on the socket
		"""

		inp = []
		outp = []
		self._fill_socketlist(inp, outp)

		(i, o, e) = select.select(inp, outp, inp, timeout)

		if len(i) + len(o) + len(e) == 0: # timeout
			return False

		self._handle_socket_events(i, o, e)
		return True

	def _request_update(self):
		PLd.requestUpdate(self.io)

	def _do(self, func, cmd):
		class handler:
			def __init__(self, pld):
				self.pld = pld
				self.ret = None
			def __call__(self, msg):
				del self.pld.handlers[cmd]
				if func:
					self.ret = func(msg)
				return True

		h = handler(self)
		self._handler_add(cmd, h)

		while self.handlers.has_key(cmd):
			if not self._process(self.timeout):
				raise IOError("Operation timed out")

		return h.ret

	def close (self):
		"""
		Immediately disconnect from packetlogic system.

		After this method has been called this object becomes
		useless.

		It may be used to force a disconnect when the garbage
		collector can't be trusted to do a timely disconnect.
		"""
		class Closed(object):
			def __getattr__(self, n):
				raise IOError("Closed")
		self.s.close()
		self.__dict__.clear()
		self.__class__ = Closed
		
	def add_update_done_callback(self, func):
		"""
		Add a callback to call after each update is done.

		@type func: C{callable}
		@param func: function to call. The function is called
		             without any arguments

		"""
		self.update_done_callbacks.append(func)

	def _request_debug_stats(self):
		pass
#		PLd.statsShowDebug(self.io)

	def _request_thread_stats(self, (zid, vid)):
		pass
#		PLd.statsShowThreads(self.io, zid, vid)

	def _request_thread_stats_all(self):
		pass
#		sysdiag=self.get_sysdiag_data()
#		for v in sysdiag.itervalues():
#			for v in v.itervalues():
#				if not (v['flags'] & 2):
#					self._request_thread_stats(v['identifier'])
#		# make sure new stats with threads are fetched
#		self._cache_stats = None

	def add_channels_callback(self, func):
		"""
		Add a callback to call with channel statistics data
		while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with a dict with the channel ids as key
		             and a inner dict as value.

		"""

		self.chans_callbacks.append(func)
		if len(self.chans_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats..
		PLd.channelsShow(self.io)
		PLd.requestUpdate(self.io)

	def add_netobj_callback(self, func, under=None, include_hosts=False):
		"""
		Add a callback to call with realtime NetObject
		data while updating.

			>>> PATH="/NetObjects/Everyone"
			>>> def cb(objs):
			...     print "#%d subscribers active under %s" % (len(objs), PATH)
			...     # print top10
			...     objs.sort(key=lambda no: no.speed[0] + no.speed[1])
			...     print "%-40s %9s %9s" % ("# Name", "inkbps", "outkbps")
			...     for no in objs[:10]:
			...         # convert to kbps
			...         inspd = no.speed[0]*8.0/1000
			...         outspd = no.speed[1]*8.0/1000
			...         print "%-40s %8.2f %8.2f" % (no.name, inspd, outspd)
			... 
			>>> rt.add_netobj_callback(cb, under=PATH)
			>>> rt.update_forever()

		@type func: C{callable}
		@param func: function to call. The function is called
		             with one argument, an L{RealtimeNetObjectTree}
			     containing the list of L{RealtimeNetObject}s data,
			     which is fetched with that objects methods.

		@type under: C{str}
		@param under: Request netobject data under a specified
		              path. In v12.0 and newer only toplevel
		              objects are included by default, and
		              this in required to get data for a
		              specific path. Observe that only the
		              objects directly under the specified
		              path is reported, no recursive
		              listing. When 'under' is used the
		              callback is called with a list of
		              L{RealtimeNetObject}s directly, no
		              L{RealtimeNetObjectTree} intermediate
		              layer.

		@type include_hosts: C{bool}
		@param include_hosts: Request information about hosts
		                      under the netobject specified
		                      with the under parameter. The
		                      argument passed to func will
		                      have an attribute named "hosts"
		                      which is an iterable of
		                      L{RealtimeHost} objects.

		@raises PLDUnsupportedInFirmware: If under argument is used and firmware doesn't support it.

		@attention: under parameter requires v12.1 firmware or newer.
		@attention: When 'under' is used the
		            callback is called with a list of
		            L{RealtimeNetObject}s directly, no
		            L{RealtimeNetObjectTree} intermediate
		            layer.
		@attention: include_hosts parameter requires v12.2 firmware or newer.
		"""
		if under is not None:
			if under[-1] != '/':
				under = under + "/"
			self.interesting_netobjects.append(under)

		self.no_callbacks.append((func, under, include_hosts))
		# Add a handler to NO_ADD messages (not done before, to prevent no-msg:s sent on commit to allocate memory, when not needed.)
		self._handler_add(PLd.PROTOCOL_SERVER_NO_ADD, self._handle_netobj_add)
		self._handler_add(PLd.PROTOCOL_SERVER_NO_REMOVE, self._handle_netobj_remove)

		if len(self.no_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats.. (For NetObjects we must also request an update to actually get any data)
		PLd.netobjectsShow(self.io)
		PLd.requestUpdate(self.io)

	def add_service_callback(self, func):
		"""
		Add a callback to call with realtime service
		data while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with one argument, a C{dict} of services.
			     Key is servicename, value is a C{dict} of its properties.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		self.service_callbacks.append(func)

		if len(self.service_callbacks) > 1:
			# No need to add at server again
			return

		# Tell server we want stats.. (For NetObjects we must also request an update to actually get any data)
		PLd.servicesShow(self.io)
		PLd.requestUpdate(self.io)

	def add_shapingobject_callback(self, func):
		"""
		Add a callback to call with realtime shapingobject
		data while updating.
		
		@type func: C{callable}
		@param func: function to call. The function is called
		             with one argument, a C{dict} with shapingobject data.
			     Key is shaping object index, value is a C{dict} of its properties.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.2 firmware or newer.
		"""
		self.shapingobject_callbacks.append(func)

		if len(self.shapingobject_callbacks) > 1:
			# No need to add at server again
			return

		PLd.shapingobjectShow(self.io)
		PLd.requestUpdate(self.io)


	def add_host_callback(self, hostip, func):
		"""
		Add a callback to be called with realtime connection information for all connections on specified host

		@type hostip: C{str}
		@param hostip: IP address of host to retrieve realtime
		               connection information for. Should be
		               an IP address as a string.

		@type func: C{callable}
		@param func: function to call. The function is called
		             with two arguments, the ipaddress and a
		             list of connections L{RealtimeConn} on
		             that host. The list is None (not to be
		             confused with the empty list) when the
		             host is removed and no more updates will
		             happen.

		@attention: This function requires v13.0 firmware or newer.
		"""
		def internal_wrapper_compatiblility_function_created_in_add_host_callback(data):
			func(hostip, data)
		vb = self.get_view_builder()
		vb.filter('Host', hostip)
		self.add_conn_view_callback(vb, internal_wrapper_compatiblility_function_created_in_add_host_callback)

	def get_view_builder(self):
		"""
		Get a ViewBuilder object for creating views.

		See L{ViewBuilder} object documentation for details.

		@rtype: L{ViewBuilder}

		@attention: This function requires v14.1 firmware or newer.
		"""
		if self._view_filters is None:
			# _fullflush will ensure that all messages from
			# server are read. PROTOCOL_SERVER_VIEW_CAPS
			# should be one of those messages causing
			# _handle_view_caps to set _view_filters and
			# _view_fields.
			self._fullflush()
		return ViewBuilder(self._view_filters, self._view_fields)

	def add_conn_view_callback(self, viewbuilder, func):
		"""
		Add a view reporting all connections matching the filter in specified viewbuilder.

		@type viewbuilder: L{ViewBuilder}
		@param viewbuilder: ViewBuilder defining the filter.
		@type func: C{callable}
		@param func: function to call. The function is called with a list of L{RealtimeConn} as its argument.

		@attention: This function requires v14.1 firmware or newer.
		"""
		if viewbuilder.fields:
			raise ValueError("No distribution allowed in add_conn_view_callback")
		PLd.createView(self.io, self.nextviewid, viewbuilder.filters, viewbuilder.fields)
		self._conn_views.append((self.nextviewid, func))
		self.nextviewid += 1

	def add_aggr_view_callback(self, viewbuilder, callback):
		"""
		Add a view reporting aggregated data.

		Data is aggregated in a tree structure according to
		the distribution and filter in the specified
		viewbuilder.

		@type viewbuilder: L{ViewBuilder}
		@param viewbuilder: ViewBuilder defining the filter and distribution.
		@type func: C{callable}
		@param func: function to call. The function is called with a L{RealtimeAggrNode} as its argument. This is the root of the aggregation tree.

		@attention: This function requires v14.1 firmware or newer.
		"""
		if not viewbuilder.fields:
			raise ValueError("No distribution required in add_conn_view_callback")
		PLd.createView(self.io, self.nextviewid, viewbuilder.filters, viewbuilder.fields)
		fieldtypes = []
		for (f, _) in viewbuilder.fields:
			fieldtypes.append([x for x in self._view_fields if x['field'] == f][0])
		self._aggr_views[self.nextviewid] = ({0: RealtimeAggrNode("Root", 0)}, fieldtypes, {0: "<None>"}, callback)
		self.nextviewid += 1

	def get_channels_data(self, maxage=5):
		"""
		Get channel statistics data once.
		
		(Should not be used simultaneously to
		add_channels_callback and update_forever)

		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		
		@rtype: C{dict}
		@returns: Dict containing all channels. Channel id is used
		          as key and the value is another dict containing
		          values.
		
		"""
		if self._cache_channels is None or time.time() >= self._cache_channels_ts + maxage:
			self.add_channels_callback(self._update_once_callback)
			self._cache_channels = self._update_once()
			self._cache_channels_ts = time.time()
		
		return self._cache_channels
		
	def get_netobj_data(self, maxage=5):
		"""
		Get Realtime NetObject data once.

		(Should not be used simultaneously to
		add_netobj_callback and update_forever)
		
		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		
		@rtype: L{RealtimeNetObjectTree}
		@returns: An object that is iterable to fetch all NetObjects,
			otherwise it has methods to fetch specific data.
		
		"""
		if time.time() >= self._netobj_ts + maxage:
			self.add_netobj_callback(self._update_once_callback)
			self._update_once()
			self._NO._link_parents()
			self._netobj_ts = time.time()
		
		return self._NO

	def get_services_data(self, maxage=5):
		"""
		Get Realtime Services data once.

		(Should not be used simultaneously to
		add_service_callback and update_forever)
		
		@type maxage: C{int}
		@param maxage: Max age in seconds of cached information.
		               If longer time has passed since last call,
			       new data will be requested from PacketLogic.
			     
		@raises PLDPermissionError: If the user lacks permission.
		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		if time.time() >= self._service_ts + maxage:
			self.add_service_callback(self._update_once_callback)
			self._update_once()
			self._service_ts = time.time()
		
		return self._services
		
	def _update_once(self):
		""" Wait for a stats/channel callback and return."""
		last = time.time()
		self.updating = True
		self._update_once_data = None
		PLd.requestUpdate(self.io)
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			if time.time() > last + timeout:
				raise IOError("Operation timed out")
			if self._process(5):
				last = time.time()
				
		return self._update_once_data

	def _update_once_callback(self, data):
		""" Callback for the _update_once function."""
		
		self._update_once_data = data
		self.stop_updating()
		
	def update_forever(self, interval=5):
		"""
		Start updating.

		@type interval: C{int}
		@param interval: Delay between updates.

		@raises PLDPermissionError: If the user lacks permission for any of the registerd callbacks

		"""

		next = time.time()
		last = time.time()
		self.updating = True
		timeout = self.timeout
		if timeout is None:
			timeout = sys.maxint
		while self.updating:
			now = time.time()
			if now > last + timeout:
				raise IOError("Operation timed out")
			if now > next:
				PLd.requestUpdate(self.io)
				next = now + interval
			if self._process(next - now):
				last = time.time()

	def stop_updating(self):
		"""
		Stop the updating and clear callback functions.

		"""
		self.updating = False
		self.stats_callbacks = []
		self.chans_callbacks = []
		self.no_callbacks = []
		self.service_callbacks = []
		self.shapingobject_callbacks = []
#		PLd.statsHide(self.io)
		PLd.channelsHide(self.io)
		PLd.netobjectsHide(self.io)
		PLd.servicesHide(self.io)
		PLd.shapingobjectHide(self.io)
		# Remove the add-handler if we have one, no need to waste time running it when until we need the data again.
		self._handler_remove(PLd.PROTOCOL_SERVER_NO_ADD)

	def fwlog_clear(self):
		"""
		Clear the firewall log.

		@returns: Nothing

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""
		PLd.fwlog_clear(self.io)
		self._fullflush()

	def _fwlog_query(self, qtype, start_or_id, end, clear=False, no_reverse=False):
		"""
		Query the firewall log for a list of entries

		@type qtype: C{int}
		@param qtype: Query type
			* 0 List of all entries matching a single rules
			* 1 List of all entries between two points in time
			* 2 List of all M entries offset at O. Suitable for access one "page" at a time.
		
		@type clear: C{bool}
		@param clear: If True, log is cleared after values are retrieved. This clears the whole log and not the selected portion. Not supported in v12.0 firmwares.

		@type no_reverse: C{bool}
		@param no_reverse: If True, the offsets are handled from oldest entry, not newest entry. i.e fwlog_query_offsets(0,10,no_reverse=True) will list 10 oldest entries, not 10 newest. Not supported in v12.0 firmwares.

		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If clear or no_reverse parameter is used on v12.0 firmware.

		@attention: clear parameter is not supported on v12.0 firmwares.
		@attention: no_reverse parameter is not supported on v12.0 firmwares.
		"""

		class LC_entry:
			def __init__(self, func):
				self.func = func
				self.lst = []
			def __call__(self, msg):
				self.lst.append(self.func(msg))
		class LC_info:
			def __init__(self, func):
				self.func = func
			def __call__(self, msg):
				self.info = self.func(msg)
				
				
		PLd.fwlog_query(self.io, qtype, start_or_id, end, clear, no_reverse)

		lc1 = LC_info(PLd.handleFWEntryStart)
		self._handler_add(PLd.PROTOCOL_SERVER_FWLOG_ENTRY_START, lc1)
		
		lc2 = LC_entry(PLd.handleFWEntry)
		self._handler_add(PLd.PROTOCOL_SERVER_FWLOG_ENTRY, lc2)

		# run until we get _END
		self._do(None, PLd.PROTOCOL_SERVER_FWLOG_ENTRY_END)

		del self.handlers[PLd.PROTOCOL_SERVER_FWLOG_ENTRY_START]
		del self.handlers[PLd.PROTOCOL_SERVER_FWLOG_ENTRY]

		return (lc1.info, lc2.lst)
		

	def fwlog_query_rule(self, id):
		"""
		Query the firewall log for a list of entries matching a specific firewall rule

			>>> info, results = rt.fwlog_query_rule(58)
			>>> print info
			{'logentries': 14054L}
			>>> print results
			[{'client': '10.0.0.2',
			  'clienthostname': '',
			  'clientport': 2985,
			  'properties': '',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '10.0.1.3',
			  'serverhostname': '',
			  'serverport': 1512,
			  'service': 'Skype',
			  'time': 1138204680,
			  'type': 0},
			 {'client': '1.2.3.4',
			  'clienthostname': '',
			  'clientport': 3247,
			  'properties': '',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '10.3.2.1',
			  'serverhostname': '',
			  'serverport': 17561,
			  'service': 'Undetermined',
			  'time': 1138204680,
			  'type': 0},
			 {'client': '3.2.1.2',
			  'clienthostname': '',
			  'clientport': 52263,
			  'properties': 'Serverinfo= AAMAILE.int.aalogistik.com Microsoft ESMTP MAIL Service, Version: 6.0.3790.1830 ready at  Wed, 25 Jan 2006 16:57:54 +0100 \\n',
			  'protocol': 6,
			  'rule': 'Monitor ',
			  'server': '1.2.4.3',
			  'serverhostname': '',
			  'serverport': 25,
			  'service': 'Undetermined',
			  'time': 1138204680,
			  'type': 0}]

		@type id: C{int}
		@param id: Rule id
		
		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{dict} and C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_ID, id, 0)

	

	def fwlog_query_offsets(self, first_nr, last_nr, clear=False, no_reverse=False):
		"""
		Query the firewall log for a list of entries between two offsets

			>>> info, results = rt.fwlog_query_offsets (0, 50)
			>>> print info
			{'logentries': 14054L}
			>>> print len(results)
			50
			>>> for i in results[0:3]:
			....:     print i
			....:
			{'clientport': 39682, 'protocol': 6, 'service': 'Lunarstorm live', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '1.2.3.4', 'client': '5.6.7.8', 'serverport': 8080, 'time': 1138204688, 'type': 0, 'properties': ''}
			{'clientport': 55172, 'protocol': 6, 'service': 'HTTP', 'clienthostname': '', 'serverhostname': 'www.google.com', 'rule': 'Log HTTP', 'server': '209.85.129.104', 'client': '192.168.9.174', 'serverport': 80, 'time': 1250775399L, 'type': 0, 'properties': 'File length=0\\nServer Hostname=www.google.com\\nURL=http://www.google.com/search?q=packetlogic&ie=utf-8&oe=utf-8&aq=t\\nRequest method=GET\\nDirection=Inbound\\nFilename=/search?q=packetlogic&ie=utf-8&oe=utf-8&aq=t\\nUser-Agent=Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.12) Gecko/2009072221\\nTransfer-Encoding=None\\nUpload file length=0\\n'}
			{'clientport': 3252, 'protocol': 6, 'service': 'eDonkey transfer', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '10.10.11.11', 'client': '40.50.60.70', 'serverport': 7804, 'time': 1138204688, 'type': 0, 'properties': ''}


		Grab all entries in the firewall log and clear once we get the last batch
		
			>>> l = []
			>>> numentries = rt.fwlog_query_offsets(0,0)[0]['logentries']
			>>> for x in range(0, numentries, 500):
			....:	l += rt.fwlog_query_offsets(x, min(x+500, numentries), no_reverse=True)[1]
			....:
			>>> l += rt.fwlog_query_offsets(numentries, -1, clear=True, no_reverse=True)[1] # grab entries logged while quering



		@type first_nr: C{str}
		@param first_nr: Start the list at this offset. 0 means the start of the firewall log. 
		
		@type last_nr: C{str}
		@param last_nr: End the list at this offset.
		
		@type clear: C{bool}
		@param clear: If True, log is cleared after values are retrieved. This clears the whole log and not the selected portion. Not supported in v12.0 firmwares.

		@type no_reverse: C{bool}
		@param no_reverse: If True, the offsets are handled from oldest entry, not newest entry. i.e fwlog_query_offsets(0,10,no_reverse=True) will list 10 oldest entries, not 10 newest. Not supported in v12.0 firmwares.

		@returns: 2-tuple with firewall log information and matching entries. last_nr - first_nr entries will be returned.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedInFirmware: If clear or no_reverse parameter is used on v12.0 firmware.

		@attention: clear parameter is not supported on v12.0 firmwares.
		@attention: no_reverse parameter is not supported on v12.0 firmwares.
		"""

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_OFFSETS, first_nr, last_nr, clear, no_reverse)

	def fwlog_query_time_range(self, timefr, timeto):
		"""
		Query the firewall log for a list between two points in time

			>>> info,results = rt.fwlog_query_time_range ("2006-01-01 00:00", "2006-02-02 00:00")
			>>> print results[0]
			{'clientport': 38742, 'protocol': 6, 'service': 'eDonkey transfer', 'clienthostname': '', 'serverhostname': '', 'rule': 'Monitor ', 'server': '3.3.3.3', 'client': '1.1.1.1', 'serverport': 4662, 'time': 1138204658, 'type': 0, 'properties': ''}

		@type timefr: C{str}
		@param timefr: Start the list at this date, the date has to be in I{YYYY-MM-DD HH:MM} format
		
		@type timeto: C{str}
		@param timeto: End the list at this date, the date has to be in I{YYYY-MM-DD HH:MM} format
			
		@returns: 2-tuple with firewall log information and matching entries.
		@rtype: C{tuple} of C{list} and C{dict}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		timefr = time.mktime(time.strptime(timefr, "%Y-%m-%d %H:%M"))
		timeto = time.mktime(time.strptime(timeto, "%Y-%m-%d %H:%M"))

		""" We have the time in UTC but we want to send it in localtime """
		timefr -= time.altzone
		timeto -= time.altzone

		return self._fwlog_query(PLd.PROTOCOL_FWRULE_QTYPE_FROMTO, long(timefr), long(timeto))

	def vcrc(self, criteria):
		"""
		Run virtual connection ruleset checker.

			>>> rt.vcrc({'serverip':'192.168.9.1'})


			>>> rt.vcrc({'serverip': '192.168.9.1',
				    'service': 'HTTP',
				    'vlan': 1234})

			>>> rt.vcrc({'serverip':'217.73.103.54'})
			{'fwrules': [(60L, 2031L), (63L, 2031L)], 'shapingrules': [(65L, 2031L),
			(66L, 1903L)], 'monitor_iface': 0L, 'fwrule': 63, 'ruleset_state': 3,
			'test_id': 1L, 'rewrite_obj': 0L}
		
		@type criteria: C{dict}
		@param criteria: Dict with any of the following keys:

		 * vlan
		 * mpls
		 * protocol
		 * clientip
		 * serverip
		 * clientport
		 * serverport
		 * service
		 * xfb
		 * properties

		@returns: dict with information, including lists for shaping and firewall rules with a tuple per matching rule. The first element in each tuple is the rule id,
		and the second element is a bitmask which describes what criteria for that rule are mutable. Mutable in this context means that there exist values for that
		criteria which will make the rule not match.
		The bitmask has the following bits defined:
		vlan = 0
		mpls = 1
		protocol = 2
		client ip = 3
		server ip = 4
		client port = 5
		server port = 6
		service = 7
		bgp = 8
		xfb = 9
		properties = 10
		The ruleset_state returned in the dict signifies the final firewall state for the connection. It is one of the following:
		UNKNOWN = 0
		ACCEPT = 1
		REJECT = 2
		DROP = 3
		REWRITE = 4
		DIVERT = 5
		"""

		PLd.vcrcTest(self.io, 1, criteria)
		return self._do(PLd.handleVcrcTest, PLd.PROTOCOL_SERVER_VCRC_TEST_RESULT)
		

	def dump_drdlstats(self):
		"""
		Dump drdl stats

		This function is just for debugging purposes. You probably
		don't know what the data means anyway :)

		@returns: the stats
		@rtype: C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		"""

		PLd.drdlStatsDump(self.io)

		lst = []

		def h(msg):
			PLd.handleDrdlStatsDump(msg, lst)
		self._handler_add(PLd.PROTOCOL_SERVER_DRDL_STATS_ENTRIES, h)

		# run until we get _DONE
		self._do(None, PLd.PROTOCOL_SERVER_DRDL_STATS_END)

		del self.handlers[PLd.PROTOCOL_SERVER_DRDL_STATS_ENTRIES]

		return lst

	def dump_bgptable(self):
		"""
		Dump the bgp table.

		This function is just for debugging purposes. It will return
		the BGP table in packetlogicd.

		@returns: tuple with all prefix and paths.
		@rtype: C{tuple} of C{list}

		@raises PLDPermissionError: If the user lacks permission for requested action
		@raises PLDUnsupportedFeatureError: If the 'bgp' feature isn't enabled.

		@attention: This API is for debugging purpose only and might change without notice.
		@attention: The API for this function changed with 14.1 to return both
		            IPv4 and IPv6 prefixes as strings. The return tuple for prefixes
		            contains (depth, prefixstr, path_id, lchildren, rchildren).
		"""
		
		self._require_features('bgp')
		
		class LC:
			def __init__(self, func):
				self.func = func
				self.lst = []
				
			def __call__(self, msg):
				self.lst.append(self.func(msg))


		PLd.bgpDump(self.io)

		lc1 = LC(PLd.handleBgpEntry)
		self._handler_add(PLd.PROTOCOL_SERVER_BGP_DUMP_ENTRY, lc1)

		lc2 = LC(PLd.handleBgpPath)
		self._handler_add(PLd.PROTOCOL_SERVER_BGP_DUMP_PATH, lc2)

		# run until we get _DONE
		self._do(None, PLd.PROTOCOL_SERVER_BGP_DUMP_DONE)

		del self.handlers[PLd.PROTOCOL_SERVER_BGP_DUMP_ENTRY]
		del self.handlers[PLd.PROTOCOL_SERVER_BGP_DUMP_PATH]

		return (lc1.lst, lc2.lst)

	def inject_packet(self, iface, pkt):
		"""
		Inject a packet.

		This function is just for testing purposes.

		@raises PLDPermissionError: If the user lacks permission for requested action

		@raises ValueError: If the size of a packet exceeds 4096 bytes.
		"""

		PLd.injectPacket(self.io, iface, pkt)
		self._fullflush()

	def force_statistics_write(self):
		"""
		Force Statistics write to the disk.

		@raises PLDUnsupportedInFirmware: If firmware doesn't support this function.

		@attention: This function requires v12.1 firmware or newer.
		"""
		PLd.force_statistics_write(self.io)
		self._fullflush()

	def _set_debug(self, facility, argument):
		PLd.setDebug(self.io, facility, argument)
		self._fullflush()



#class StatsZone(PLd.ClientStatsZone):
#	"""Class for registering system diagnostics zone.
#
#			>>> zone = pld.StatsZone('TestStats')
#			>>> zone.add_size('Entries')
#			>>> zone.add_timespan('Uptime')
#			>>> zone.register(r)
#			>>> while True:
#			...     zone['Entries'] = 1337
#			...     zone['Uptime'] = time.time() - starttime
#			...     zone.send(r)
#			...     time.sleep(3)
#
#		@raises KeyError: If an invalid key is specified using [ ] index.
#	"""
#	
#
#	def __init__(self, name, idx=0, oid="", hash_oid=True):
#		"""
#
#		@type name: C{str}
#		@param name: Name of the system diagnostics zone to registered
#		
#		@type idx: C{int}
#		@param idx: Zone index, used if multiple zones are to be registered. Only available in v12.1 firmware and newer.
#
#		@type oid: C{str}
#		@param oid: Register this zone with specified OID for SNMP. Only available in v12.4 firmware and newer.
#
#		@type hash_oid: C{bool}
#		@param hash_oid: Automatically add a unique index at end of this OID. Only available in v12.4 firmware and newer.
#
#		@raises PLDUnsupportedInFirmware: If idx is used without firmware support.
#
#		@attention: Registering multiple zones requires v12.1 firmware or newer.
#		@attention: Registering zones with oid requires v12.4 firmware or newer.
#		"""
#		PLd.ClientStatsZone.__init__(self, name, idx, oid, hash_oid)
#		self.values = {}
#		
#	def register(self, pld):
#		"""
#		Register zone to PLd connection.
#		"""
#		PLd.ClientStatsZone.register(self, pld.io)
#		pld._flush()
#
#	def send(self, pld):
#		"""
#		
#		"""
#		PLd.ClientStatsZone.send(self, pld.io)
#		pld._flush()
#		
#	def add(self, name, typ):
#		"""
#		@raises RuntimeError: if this StatsZone has been registered already
#		
#		@raises ValueError: if a value with the given name is added already.
#		"""
#		PLd.ClientStatsZone.add(self, name, typ)
#		self.values[name] = None
#
#	def add_interval(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_INTERVAL)
#
#	def add_moment(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_MOMENT)
#
#	def add_size(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_SIZE)
#
#	def add_speed(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_SPEED)
#
#	def add_avg(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_AVG)
#
#	def add_timespan(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_TIMESPAN)
#
#	def add_percent(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_PERCENT)
#
#	def add_time(self, name):
#		self.add(name, PLd.SYSDIAG_TYPE_TIME)
#
#	def __setitem__(self, name, value):
#		if self.values.has_key(name):
#			self.set(name, value)
#			self.values[name] = value
#			return
#		raise KeyError, name
#	
#	def __getitem__(self, name):
#		if self.values.has_key(name):
#			return self.values[name]
#		raise KeyError, name

class RealtimeAggrNode(object):
	"""
	This object should not be instantiated manually. Instances
	will be passed as argument to the function registered with
	L{PLD.add_aggr_view_callback}.
	"""
	def __init__(self, name, parentid):
		self._name = name
		self.parentid = parentid
		self._children = []
		self.data = {}

	def _addchild(self, c):
		self._children.append(c)
	def _removechild(self, c):
		self._children.remove(c)
	def _set_data(self, data):
		self.data.update(data)

	@property
	def name(self):
		"""
		The name of this node
                """
		return self._name

	@property
	def connections(self):
		"""
		Number of connections
		"""
		return self.data.get("conns")

	@property
	def speed(self):
		"""
		inbound/outbound bps
		"""
		return self.data.get("speed")

	@property
	def children(self):
		"""
		List of child nodes (if any).
		"""
		return self._children

class ViewBuilder(object):
	"""
	Builder for creating views with filters and distributions.

	This object should not be instantiated manually. Use the
	L{PLD.get_view_builder} to create a new object.
	"""

	def __init__(self, filtermeta, fieldmeta):
		self.filtermeta = filtermeta
		self.fieldmeta = fieldmeta
		self.fields = []
		self.filters = []

	def _get_filtermeta(self, k):
		for e in self.filtermeta:
			if e['name'] == k:
				return e
		raise ValueError("Unknown filter key %r" % k)

	def _get_fieldmeta(self, k):
		for e in self.fieldmeta:
			if e['name'] == k:
				return e
		raise ValueError("Unknown field %r" % k)

	def _filter_on(self, key, neg, values):
		meta = self._get_filtermeta(key)

		if meta['type'] == PLd.PROTOCOL_VIEW_FILTER_TYPE_SMALL_INTS:
			rtyp = int
		elif meta['type'] == PLd.PROTOCOL_VIEW_FILTER_TYPE_HOSTS:
			rtyp = basestring
		elif meta['type'] in (PLd.PROTOCOL_VIEW_FILTER_TYPE_SERVICES,
				      PLd.PROTOCOL_VIEW_FILTER_TYPE_OBJECTS,
				      PLd.PROTOCOL_VIEW_FILTER_TYPE_STRINGS):
			rtyp = basestring
		else:
			raise RuntimeError("Unknown filter type")

		if len(values) == 0:
			raise ValueError("At least one value needed")

		for f, _, _, _ in self.filters:
			if f == meta['filter']:
				raise ValueError("Filter of that type already added")

		for v in values:
			if not isinstance(v, rtyp):
				raise TypeError("Value not of correct type")

		self.filters.append((meta['filter'], neg, meta['type'], values))

	def filter(self, key, *values):
		"""
		Add a filter to the view.

		Makes filter only include connections where the
		property specified by key is one of the specified
		values.

			>>> v = rt.get_view_builder()
			>>> v.filter("Host", "10.0.0.1")
			>>> v.filter("Server Port", 443)

		@type key: C{str}
		@param key: The key to filter on.

		"""
		self._filter_on(key, False, values)

	def filter_out(self, key, *values):
		"""
		Add a filter to exclude traffic from the view.

		Makes filter exclude connections where the
		property specified by key is one of the specified
		values.

			>>> v = rt.get_view_builder()
			>>> v.filter("Host", "10.0.0.1")
			>>> v.filter("Server Port", 443)
			>>> v.filter_out("Service", "SSL v3")

		@type key: C{str}
		@param key: The key to filter on.

		"""
		self._filter_on(key, True, values)

	def distribution(self, field, arg=None):
		"""
		Add a distribution to the view.

		Adds the specified field as a sub-distribution
		of current distribution.

			>>> v = rt.get_view_builder()
			>>> v.filter_out("Service", "QUIC")
			>>> v.distribution("Server Hostname")
			>>> v.distribution("NetObject Level", "DHCP/By MAC")
			>>> v.distribution("Local Host")


		@type field: C{str}
		@param field: The field to distribute based on.
		@type arg: C{str}
		@param arg: Argument as required by some fields
		"""
		meta = self._get_fieldmeta(field)

		if meta['want_arg']:
			if arg is None:
				raise ValueError("Field %r requires argument" % field)
			if not isinstance(arg, basestring):
				raise TypeError("Argument must be string")
		elif arg is not None:
			raise ValueError("Field %r can't take argument" % field)
		else:
			arg = ""
		self.fields.append((meta['field'], arg))

	@property
	def available_filters(self):
		"""
		List of available filter keys.
		"""
		return [f['name'] for f in self.filtermeta]

	@property
	def available_fields(self):
		"""
		List of fields available for use in distributions.
		"""
		return [f['name'] for f in self.fieldmeta]

	def dump_meta(self):
		"""
		Print a human readable representation of available filters and fields to stdout.
		"""
		print "Available filters:"
		for f in self.filtermeta:
			print " ", f['name'],
			if f['type'] == PLd.PROTOCOL_VIEW_FILTER_TYPE_SMALL_INTS:
				print "(number (max %d))" % f['max']
			elif f['type'] == PLd.PROTOCOL_VIEW_FILTER_TYPE_HOSTS:
				print "(ipaddress)"
			elif f['type'] in (PLd.PROTOCOL_VIEW_FILTER_TYPE_SERVICES,
					   PLd.PROTOCOL_VIEW_FILTER_TYPE_OBJECTS,
					   PLd.PROTOCOL_VIEW_FILTER_TYPE_STRINGS):
				print "(string)"
			else:
				print "(???)"

		print "Available fields for distribution:"
		for f in self.fieldmeta:
			if f['want_arg']:
				print " ", f['name'], " (needs argument)"
			else:
				print " ", f['name']


class RealtimeConn(object):
	"""
	Object containing realtime information of a connection. This object
	should not be instantiated manually. Use the
	L{add_host_callback<PLD.add_host_callback>} method.


	        >>> def cb(ip, connections):
	        ...    print "=" * 30
	        ...    print time.ctime()
	        ...    print ip
	        ...    if connections is None:
	        ...        print " * Host removed"
	        ...        rt.stop_updating()
	        ...    else:
	        ...        for c in connections:
	        ...            if c.service != "HTTP":
	        ...                continue
	        ...            f = " * "
	        ...            f += "%s:%d" % c.client
	        ...            f += "->%s:%d" % c.server
	        ...            if c.server_hostname:
	        ...                f += "[%r]" % c.server_hostname
	        ...            f += " in=%dbps out=%dbps" % c.speed
	        ...            f += " prio=%d" % c.shaping_prio
	        ...            print f
	        >>> rt.add_host_callback(host, cb)
	        >>> rt.update_forever(5.0)
	"""
	def __init__(self, connid, protocol, clientip, clientport, serverip, serverport):
		self._connid = connid
		self._protocol = protocol
		self._clientip = clientip
		self._clientport = clientport
		self._serverip = serverip
		self._serverport = serverport
		self._starttime = 0
		self._server_hostname = None
		self._speed = (0,0)
		self._xfer = (0,0)
		self._shaping_prio = None
		self._shaping_rule_ids = []
		self._stat_rule_ids = []
		self._views = []

	def _set_views(self, newviews):
		self._views = newviews

	def _in_view(self, view):
		return view in self._views

	speed = property(lambda s: (s._speed[0] * 8, s._speed[1] * 8),
	                 doc="Average speed of the connection during the last update interval as a tuple (inbound bps, outbound bps)")
	xfer = property(lambda s: s._xfer,
	                doc="Bytes transferred during the lifetime of the connection as a tuple (inbound bytes, outbound bytes)")
	protocol = property(lambda s: s._protocol,
	                    doc="IP protocol used by connection")
	client = property(lambda s: (s._clientip, s._clientport),
	                  doc="Client endpoint of the connection as a tuple (clientip, clientport)")
	server = property(lambda s: (s._serverip, s._serverport),
	                  doc="Server endpoint of the connection as a tuple (serverip, serverport)")
	service = property(lambda s: s._service,
	                   doc="Service of the connection as a string")
	base_service = property(lambda s: s._base_service,
	                   doc="Base Service of the connection as a string")
	server_hostname = property(lambda s: s._server_hostname,
	                           doc="Server hostname if extracted by signatures")
	shaping_prio = property(lambda s: s._shaping_prio,
	                        doc="Priority value of the connection as assigned by shaping rules")
	shaping_rules = property(lambda s: s._shaping_rules,
	                         doc="Shaping rules matching the connection as a list of rule names")
	starttime = property(lambda s: s._starttime,
	                     doc="Unix timestamp when first packet in connection was seen (according to the clock on the PacketLogic system)")

def _calc_qoe(packets, issues):
	if packets == 0:
		return -1
	if issues == 0:
		return 100
	return 100 - int(100.0 * issues / packets + 0.5)

class RealtimeHost(object):
	"""
	Object containing realtime information of a host. This object
	should not be instantiated manually. Use the
	L{add_netobj_callback<PLD.add_netobj_callback>} method.

	        >>> def cb(data):
	        ...     print "Number of hosts: %d" % len(data.hosts)
	        ...     interesting_hosts = ["10.2.200.188", "10.2.20.58"]
	        ...     for h in data.hosts:
	        ...         if h in interesting_hosts:
	        ...             print " %s: %s" % (h.addr, h.speed)
                ...
	        >>> rt.add_netobj_callback(cb, under="/NetObjects/<Ungrouped>", include_hosts=True)
	        >>> rt.update_forever()                     


	@type addr: C{str}
	@ivar addr: IP address of host

	@type speed: C{tuple} of two C{int}s
	@ivar speed: Speed in byte/s in/out.
	
	@type connections: C{tuple} of two C{int}s
	@ivar connections: Connections all/unestablished.
	
	@type cps: C{tuple} of two C{int}s
	@ivar cps: Connections per second in/out.

	@type connprot: C{bool}
	@ivar connprot: Host has hit connection protection limit.
	"""
	def __init__(self, hid):
		"""
		This class is instantiated internally, no need to instantiate it manually.
		"""
		self.id = hid
		self.addr = None
		self.speed = (-1, -1)
		self.connections = (-1, -1)
		self.cps = (-1, -1)
		self.connprot = False

		self.quality_packets = (-1, -1)
		self.quality_internal_issues = (-1, -1)
		self.quality_external_issues = (-1, -1)


	def _inbound_internal_quality(self):
		return _calc_qoe(self.quality_packets[0], self.quality_internal_issues[0])
	def _inbound_external_quality(self):
		return _calc_qoe(self.quality_packets[0], self.quality_external_issues[0])
	def _outbound_internal_quality(self):
		return _calc_qoe(self.quality_packets[1], self.quality_internal_issues[1])
	def _outbound_external_quality(self):
		return _calc_qoe(self.quality_packets[1], self.quality_external_issues[1])

	internal_quality = property(lambda s: (s._inbound_internal_quality(), s._outbound_internal_quality()),
				    doc="""Internal QoE values (inbound, outbound).

                                        Internal Quality of Experience
                                        values in range 0-100
                                        (percent) for this host. Value
                                        is -1 if not available.

		                        @raises PLDUnsupportedInFirmware: If accessed without firmware support.

		                        @attention: Quality values requires firmware 13.1 or newer.
                                        """)
	external_quality = property(lambda s: (s._inbound_external_quality(), s._outbound_external_quality()),
				    doc="""External QoE values (inbound, outbound).

                                        External Quality of Experience
                                        values in range 0-100
                                        (percent) for this host. Value
                                        is -1 if not available.

		                        @raises PLDUnsupportedInFirmware: If accessed without firmware support.

		                        @attention: Quality values requires firmware 13.1 or newer.
                                        """)

	def __eq__(self, other):
		if isinstance(other, basestring):
			return self.addr == other
		if isinstance(other, RealtimeHost):
			return self.addr == other.addr
		return False

	def __ne__(self, other):
		return not self.__eq__(other)

	def __str__(self):
		return "<RealtimeHost %s>" % self.addr
	__repr__ = __str__

class RealtimeNetObject(object):
	"""
	Object containing Realtime NetObject data.

	@type id: C{int}
	@ivar id: The internal id.
	
	@type parent: L{RealtimeNetObject} or C{None}
	@ivar parent: References the parent NetObject, if any.
	
	@type name: C{str}
	@ivar name: Name of NetObject.
	
	@type speed: C{tuple} of two C{int}s
	@ivar speed: Speed in byte/s in/out.
	
	@type forwarded: C{tuple} of two C{int}s
	@ivar forwarded: Forwarded bytes in/out. Obsolete in v12.0 and newer, will always be 0 in these versions.
	
	@type connections: C{tuple} of two C{int}s
	@ivar connections: Connections in/out.
	
	@type cps: C{tuple} of two C{int}s
	@ivar cps: Connections per second in/out.
	
	"""
	#__slots__ = ('id', 'parent', 'path', 'name', 'speed', 'forwarded', 'connections', 'cps')
	__slots__ = ('id', 'parent', 'fullpath', 'name', 'speed', 'forwarded', 'connections', 'cps')
	
	def __init__(self):
		"""
		This class is instantiated internally, no need to instantiate it manually.
		"""
		self.id = 0L
		self.parent = 0L
		self.name = ''
		self.speed = (0L,0L)
		self.forwarded = (0L,0L)
		self.connections = (0L,0L)
		self.cps = (0L, 0L)

	def __repr__(self):
		return "<%s '%s' %d>" % (self.__class__.__name__, self.name, self.id)

	fullpath = property(lambda self: "%s/%s" % (self.path, self.name), doc="Full path of this object including the object's name.")
	def _getpath(self):
		if not self.parent:
			return '/NetObjects'
		else:
			return self.parent.fullpath
	
	path = property(_getpath, doc="Path which this object resides under.")

class _RealtimeNetObjectList(list):
	def _hosts_getter(self):
		if not hasattr(self, "_hosts_func"):
			return []
		return self._hosts_func()

	hosts = property(_hosts_getter)

class RealtimeNetObjectTree(object):
	"""
	NetObject data accessed through the NetObject callback function

		>>> import packetlogic2
		>>> pl = packetlogic2.connect('192.168.1.25', 'admin', 'pldemo00')
		>>> rt = pl.Realtime()
		>>> count = 1
		>>> def f(data):
		...	global count
		...	
		...	if count > 100:
		...		rt.stop_updating()
		...	if data:
		...		for m in data.list('/NetObjects/DSL'):
		...			print "name: %s" % m.fullpath
		...			print "speed: ", m.speed
		...			print "connections: ", m.connections
		...		m = data.find('/NetObjects/DSL/Lab')
		...		if m is not None:
		...			print "name: %s" % m.fullpath
		...			print "speed: ", m.speed
		...			print "connections: ", m.connections
		...	count +=1
		...
		>>> rt.add_netobj_callback(f)
		>>> rt.update_forever()			

	"""
	__slots__ = ('_cache', '_by_id')
	
	def __init__(self):
		self._clear()
		
	def _clear(self):
		self._cache = []
		self._by_id = {}

	def __len__(self):
		return len(self._cache)

	def _append(self, netobj):
		self._cache.append(netobj)
		self._by_id[netobj.id] = netobj
		
	def _extend(self, netobj_list):
		for netobj in netobj_list:
			self._append(netobj)

	def _removeobj(self, no):
		for x in self._cache[:]:
			if x.parent is no:
				self._removeobj(x)
		self._cache.remove(no)
		del self._by_id[no.id]

	def _remove(self, noid):
		no = self._by_id.get(noid)
		if no is None:
			print "Remove of nonexistant object?"
			return
		assert no.id == noid
		self._removeobj(no)
		
	def find(self, path):
		"""
		Find a single Realtime NetObject with the given full path.
		
		@returns: The specified NetObject data, or None if not found.
		@rtype: L{RealtimeNetObject}
		"""
		for obj in self._cache:
			if obj.fullpath == path:
				return obj
		return None

	#def getlist(self, path='/NetObjects'):
	#	if self._cache and isinstance(self._cache[0].parent, (int, long)):
	#		self._link_parents()
	#	return [obj for obj in self._cache if obj.getfullpath.startswith(path + "/") or obj.getfullpath == path]
			
	def list(self, path='/NetObjects', recursive=True):
		"""
		Return a list of Realtime NetObject data, matching the
		given path, and optional recursive flag.

		@type path: C{str}
		@param path: Path where to start looking. If the path
			ends with a slash (in recursive mode), it will
			include the children of that NetObject (if any)
			but not the NetObject itself.
			The path should begin with '/NetObjects'.

		@type recursive: C{bool}
		@param recursive: If True any level of children NetObjects will
			also be returned. Otherwise only the NetObjects
			that is direct children to the given full path
			is returned.
			(Do not end path with / if recursive is True).
		
		@returns: List of matching Realtime NetObjects.
		@rtype: C{list} of L{RealtimeNetObject}s
		"""
		if recursive:
			if path.endswith('/'):
				return [obj for obj in self._cache if obj.fullpath.startswith(path)]
			else:
				return [obj for obj in self._cache if obj.fullpath.startswith(path + "/") or obj.fullpath == path]
		else:
			return _RealtimeNetObjectList([obj for obj in self._cache if obj.path == path])

	def __iter__(self):
		return self._cache.__iter__()
	
	def _link_parent_one(self, netobj):
		if netobj.parent == 0:
			netobj.parent = None
		elif isinstance(netobj.parent, RealtimeNetObject):
			pass
		else:
			netobj.parent = self._by_id.get(netobj.parent, netobj.parent)

	def _link_parents(self):
		"""
		Replace property parent with that object, not the id of it.
		Root NetObjects will have parent=None.
		"""
		for netobj in self._cache:
			self._link_parent_one(netobj)
