"""
This module contains functionality for connection protection triggers.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> u = pl.ConnprotTriggers()

"""

from resource import Resource
from thisversion import PLDB

import plutil


_Prop_ConnprotTrigger = plutil.Prop('id', 'name', 'script')
class ConnprotTrigger(object):
    """
    Connection Protection triggers

    Returned from L{ConnprotTriggers.list} and L{ConnprotTriggers.add}.
    """
    __slots__ = ('_prop', '_pldb')

    def __init__(self, _pldb=None, **props):
        """
        Should not be instantiated manually, use L{list <ConnprotTriggers.list>}.
        """
        self._pldb = _pldb
        self._prop = _Prop_ConnprotTrigger(**props)

    def __repr__(self):
        return "<ConnprotTrigger '%s' (#%d) script=%r>" % (self.name, self.id, self.script)

    def update(self, script):
        """
        Modify this connection protection trigger.

        @type script: C{str}
        @param script: Script name this connection protection trigger should execute.
        """
        plutil._do(self._pldb, self._pldb.session.connprot_trigger_update(self._prop.id, script), plutil._AddCB(self._pldb))

        self._prop.script = script

    def remove(self):
        """
        Remove this connection protection trigger from the system.
        """
        plutil._do(self._pldb, self._pldb.session.connprot_trigger_remove(self._prop.id), plutil._AddCB(self._pldb))

    id = property(lambda self: self._prop.id, doc="ID Number.")
    name = property(lambda self: self._prop.name, doc="Name ")
    script = property(lambda self: self._prop.script, doc="script")


class ConnprotTriggers(Resource):
    """
    This is the resource for manipulating connection protection triggers.

    @group ConnprotTriggers: list, add
    """

    def _resource(self):
        return PLDB.RESOURCE_CONNPROT_TRIGGER

    def list(self):
        """
        Lists all connection protection triggers on the system

            >>> u.list()
            [<ConnprotTrigger 'My Trigger' (#3) script='add_to_netobject'>]

        @rtype: C{list} of L{ConnprotTrigger}
        @returns: A list of all connection protection triggers configured on system
        """
        r = plutil._do(self, self.session.connprot_trigger_list(),
		       plutil._ListCB(self, self.session.connprot_trigger_list_handle, ConnprotTrigger, False))
	return [ConnprotTrigger(_pldb=self, **x) for x in r]

    def add(self, name, script):
        """
        Adds a new connection protection trigger to the system.

        @type name: C{str}
        @param name: name of the new connection protection trigger

        @type script: C{str}
        @param script: Script name this connection protection trigger should execute.

        @rtype: L{ConnprotTrigger}
        @returns: A L{ConnprotTrigger} object representing the newly created connection protection trigger.
        """
        _id = plutil._do(self, self.session.connprot_trigger_add(name, script), plutil._AddCB(self, self.session.connprot_trigger_add_handle))

        return ConnprotTrigger(id=_id, name=name, script=script, _pldb=self)
